from naas.types import t_notebook, t_scheduler, t_asset
from .proxy import encode_proxy_url
from .env_var import n_env
from bs4 import BeautifulSoup
import pandas as pd
import pretty_cron
import requests
import base64
import json
import uuid
import os


class Notifications:
    logger = None

    headers = None

    def __init__(self, logger=None):
        self.headers = {"Authorization": f"token {n_env.token}"}
        self.logger = logger

    def send(self, email_to, subject, html, files=[], email_from=None):
        uid = str(uuid.uuid4())
        soup = BeautifulSoup(html, features="html5lib")
        content = soup.get_text()
        try:
            data = {
                "subject": subject,
                "from": email_from,
                "email": ",".join(email_to) if isinstance(email_to, list) else email_to,
                "content": content,
                "html": html,
            }
            req = None
            if len(files) > 0:
                files_list = []
                for file in files:
                    abs_path = os.path.abspath(os.path.join(os.getcwd(), file))
                    try:
                        files_list.append((file, open(abs_path, "rb")))
                    except Exception as err:
                        print(err)
                req = requests.post(
                    url=f"{n_env.notif_api}/send",
                    files=files_list,
                    headers=self.headers,
                    data=data,
                )
            else:
                req = requests.post(
                    url=f"{n_env.notif_api}/send", headers=self.headers, json=data
                )
            req.raise_for_status()
            print("👌 💌 Email has been sent successfully !")
        except Exception as err:
            if self.logger is not None:
                self.logger.error(
                    json.dumps({"id": uid, "type": "email error", "error": str(err)})
                )
            else:
                print(err)

    def send_status(
        self,
        uid,
        status,
        email_to,
        file_path,
        current_type,
        current_value,
        files=[],
        email_from=None,
    ):
        if n_env.notif_api is None:
            jsn = {"id": uid, "type": "notification error", "error": "not configured"}
            if self.logger is not None:
                self.logger.error(json.dumps(jsn))
            else:
                print(jsn)
            return jsn
        content = ""
        file_link = f"{n_env.hub_api}/user/{n_env.user}/tree/{file_path}"
        if current_type == t_asset or current_type == t_notebook:
            content = f'The file <a href="{file_link}">{file_path}</a> <br/>'
            content += f"Accesible at this url:<br/> {encode_proxy_url(current_type)}/{current_value}<br/>"
        elif current_type == t_scheduler:
            cron_string = pretty_cron.prettify_cron(current_value)
            content = f'The file <a href="{file_link}">{file_path}</a><br/>'
            content += f"who run {cron_string}<br/>"
        content += f"Is {status}.<br/><br/>"
        content += "Check the Logs on your manager below :<br/>"
        status_url = f"{encode_proxy_url(t_asset)}/naas_{status}.png"
        message_bytes = file_path.encode("ascii")
        base64_bytes = base64.b64encode(message_bytes)
        file_path_base64 = base64_bytes.decode("ascii")
        link_url = f"{n_env.hub_api}/user/{n_env.user}/naas/?filter={file_path_base64}"
        logo_url = f"{encode_proxy_url(t_asset)}/naas_logo.png"
        try:
            data = {
                "title": "Naas manager notification",
                "subject": f"{current_type.capitalize()} {status}",
                "from": email_from,
                "email": ",".join(email_to) if isinstance(email_to, list) else email_to,
                "content": content,
                "custom_vars": {
                    "URL_HOME": encode_proxy_url(),
                    "URL_LOGO": logo_url,
                    "ALT_LOGO": "Naas Manager LOGO",
                    "URL_IMAGE": status_url,
                    "ALT_IMAGE": status,
                    "URL_LINK": link_url,
                },
            }
            req = None
            if len(files) > 0:
                files = {}
                for file in files:
                    abs_path = os.path.abspath(os.path.join(os.getcwd(), file))
                    try:
                        files[file] = open(abs_path, "rb")
                    except Exception as err:
                        print(err)
                req = requests.post(
                    url=f"{n_env.notif_api}/send_status",
                    headers=self.headers,
                    files=files,
                    json=data,
                )
            else:
                req = requests.post(
                    url=f"{n_env.notif_api}/send_status",
                    headers=self.headers,
                    json=data,
                )
            req.raise_for_status()
            jsn = req.json()
            return jsn
        except Exception as err:
            if self.logger is not None:
                self.logger.error(
                    json.dumps(
                        {"id": uid, "type": "notification error", "error": str(err)}
                    )
                )
            else:
                print(err)

    def status(self):
        req = requests.get(url=f"{n_env.notif_api}/")
        req.raise_for_status()
        jsn = req.json()
        return jsn

    def list(self):
        req = requests.post(
            url=f"{n_env.notif_api}/list",
            headers=self.headers,
        )
        jsn = req.json()
        return pd.DataFrame(data=jsn.get("emails"))

    def list_all(self):
        req = requests.post(
            url=f"{n_env.notif_api}/list_all",
            headers=self.headers,
        )
        jsn = req.json()
        return pd.DataFrame(data=jsn.get("emails"))



from IPython.core.display import display, HTML
from .runner.proxy import encode_proxy_url
from .types import t_delete, t_job, t_add, t_env
from .runner.env_var import n_env
import ipywidgets as widgets
import pandas as pd
import ipykernel
import requests
import copy
import uuid
import os
import traceback
import base64


enterprise_gateway = False
try:
    import enterprise_gateway.services.kernels.remotemanager  # noqa: F401

    enterprise_gateway = True  # noqa: F811
except ImportError:
    pass


class Manager:
    __error_busy = "Naas look busy, try to reload your machine"
    __error_reject = "Naas refused your request, reason :"
    __filetype = None
    headers = None

    def __init__(self, filetype):
        self.headers = {"Authorization": f"token {n_env.token}"}
        self.__filetype = filetype
        self.set_runner_mode()

    def is_production(self):
        if "naas_env" in globals():
            return True if naas_env == "PRODUCTION" else False  # type: ignore  # noqa: F821
        else:
            return False

    def set_runner_mode(self):
        try:
            r = requests.get(
                f"{n_env.api}/{t_env}",
            )
            r.raise_for_status()
        except Exception:
            n_env.remote_mode = not n_env.remote_mode

    def get(self):
        public_url = f"{encode_proxy_url()}"
        print("You can check your current tasks list here :")
        display(HTML(f'<a href="{public_url}"">Manager</a>'))

    def get_logs(self):
        req = requests.get(url=f"{n_env.api}/logs")
        req.raise_for_status()
        jsn = req.json()
        return jsn

    def get_naas(self):
        naas_data = []
        try:
            r = requests.get(f"{n_env.api}/{t_job}")
            r.raise_for_status()
            naas_data = r.json()
        except requests.exceptions.ConnectionError:
            print(self.__error_busy)
        except requests.HTTPError as e:
            print(self.__error_reject, e)
        return naas_data

    def get_value(self, path, obj_type):
        json_data = self.get_naas()
        value = None
        for item in json_data:
            if item["type"] == obj_type and item["path"] == path:
                value = item["value"]
        return value

    def notebook_path(self):
        try:
            connection_file = os.path.basename(ipykernel.get_connection_file())
            kernel_id = connection_file.split("-", 1)[1].split(".")[0]
            if enterprise_gateway:
                kernel_id = connection_file.split("-", 1)[1].split("_")[0]
            notebooks = self.running_notebooks()
            for notebook in notebooks:
                if kernel_id in notebook["kernel_id"]:
                    return os.path.join(n_env.server_root, notebook["path"])
        except Exception as e:
            tb = traceback.format_exc()
            print("notebook_path", e, tb)
        return None

    def running_notebooks(self):
        try:
            base_url = f"{n_env.hub_api}/user/{n_env.user}/api/sessions"
            req = requests.get(url=base_url, headers=self.headers)
            req.raise_for_status()
            sessions = req.json()
            notebooks = [
                {
                    "kernel_id": notebook["kernel"]["id"],
                    "path": notebook["notebook"]["path"],
                }
                for notebook in sessions
            ]
            return notebooks
        except requests.exceptions.ConnectionError as e:
            print(e)
        except requests.HTTPError as e:
            print(e)
        except Exception as e:
            tb = traceback.format_exc()
            print("running_notebooks", e, tb)
        return []

    def get_path(self, path):
        if path is not None:
            return os.path.abspath(os.path.join(os.getcwd(), path))
        else:
            return self.notebook_path()

    def copy_clipboard(self, text):
        uid = uuid.uuid4().hex
        js = """<script>
        function copyToClipboard_{uid}(text) {
            const dummy = document.createElement("textarea");
            document.body.appendChild(dummy);
            dummy.value = text;
            dummy.select();
            document.execCommand("copy");
            document.body.removeChild(dummy);
        }
        </script>"""
        js = js.replace("{uid}", uid)
        display(HTML(js))
        js2 = f"<script>copyToClipboard_{uid}(`" + text + "`);</script>"
        display(HTML(js2))

    def copy_url(self, text):
        button = widgets.Button(description="Copy URL", button_style="primary")
        output = widgets.Output()

        def on_button_clicked(b):
            with output:
                self.copy_clipboard(text)
                html_div = '<div id="pasting_to_clipboard">✅ Copied !</div>'
                display(HTML(html_div))

        button.on_click(on_button_clicked)
        display(button, output)

    def proxy_url(self, endpoint, token=None):
        public_url = encode_proxy_url(endpoint)
        if token:
            public_url = f"{public_url}/{token}"
        return public_url

    def __open_file(self, path):
        filename = os.path.basename(path)
        if not os.path.exists(path):
            raise FileNotFoundError(f"file doesn't exist {path}")
        data = open(path, "rb").read()
        encoded = base64.b64encode(data)
        return {"filename": filename, "data": encoded.decode("ascii")}

    def safe_filepath(self, path):
        path = os.path.join(n_env.server_root, path)
        return path

    def __save_file(self, path, filedata=None):
        dirname = os.path.dirname(path)
        filename = filedata["filename"] or os.path.basename(path)
        new_path = os.path.join(dirname, filename)
        if not filedata:
            return
        if os.path.exists(new_path):
            raise FileNotFoundError(f"file already exist {new_path}")
        f = open(new_path, "wb")
        decoded = base64.b64decode(filedata["data"])
        f.write(decoded)
        f.close()

    def clear_file(self, path=None, mode=None, histo=None):
        if not path and self.is_production():
            print("No clear_prod done you are in production\n")
            return
        prod_path = self.get_path(path)
        try:
            r = requests.delete(
                f"{n_env.api}/{t_job}",
                params={"path": prod_path, "histo": histo, "mode": mode},
            )
            r.raise_for_status()
            res = r.json()
            for ff in res:
                print(f"🕣 Your Notebook output {ff} has been remove from production.\n")
            return pd.DataFrame(data=res.get("files", []))
        except requests.exceptions.ConnectionError as err:
            print(self.__error_busy, err)
            raise
        except requests.HTTPError as err:
            print(self.__error_reject, err)
            raise

    def list_prod(self, mode, path=None):
        if not path and self.is_production():
            print("No list_prod done you are in production\n")
            return
        current_file = self.get_path(path)
        try:
            r = requests.get(
                f"{n_env.api}/{t_job}",
                params={"path": current_file, "type": self.__filetype, "mode": mode},
            )
            r.raise_for_status()
            res = r.json()
            print(
                "\n+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++\n",
                res,
                "\n+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++\n",
            )
            if res.get("files", None) and len(res.get("files", [])) > 0:
                return pd.DataFrame(data=res.get("files", []))
            else:
                print("No files found in prod")
                return []
        except requests.exceptions.ConnectionError as err:
            print(self.__error_busy, err)
            raise
        except requests.HTTPError as err:
            print(self.__error_reject, err)
            raise

    def get_file(self, path=None, mode=None, histo=None):
        if not path and self.is_production():
            print("No get_prod done you are in production\n")
            return
        current_file = self.get_path(path)
        filename = os.path.basename(current_file)
        try:
            r = requests.get(
                f"{n_env.api}/{t_job}",
                params={
                    "path": current_file,
                    "type": self.__filetype,
                    "mode": mode,
                    "histo": histo,
                },
            )
            r.raise_for_status()
            res = r.json()
            self.__save_file(self.safe_filepath(current_file), res.get("file"))
            print(
                f"🕣 Your Notebook {mode or ''} {filename}, has been copied into your local folder.\n"
            )
        except requests.exceptions.ConnectionError as err:
            print(self.__error_busy, err)
            raise
        except requests.HTTPError as err:
            print(self.__error_reject, err)
            raise

    def path(self, filetype):
        def mode_path(self, path):
            nonlocal filetype
            if self.is_production():
                # filename = os.path.basename(path)
                # dirname = os.path.dirname(path)
                # filename = f"{filetype}_{filename}"
                # type_path = os.path.join(dirname, filename)
                type_path = path
                return self.get_path(type_path)
            else:
                return path

        return mode_path

    def add_prod(self, obj, debug):
        if "type" in obj and "path" in obj and "params" in obj and "value" in obj:
            new_obj = copy.copy(obj)
            dev_path = obj.get("path")
            new_obj["path"] = self.get_path(dev_path)
            new_obj["file"] = self.__open_file(dev_path)
            new_obj["status"] = t_add
            try:
                if debug:
                    print(f'{new_obj["status"]} ==> {new_obj}')
                r = requests.post(f"{n_env.api}/{t_job}", json=new_obj)
                r.raise_for_status()
                res = r.json()
                if debug:
                    print(f'{res["status"]} ==> {res}')
            except requests.exceptions.ConnectionError as err:
                print(self.__error_busy, err)
                raise
            except requests.HTTPError as err:
                print(self.__error_reject, err)
                raise
            return new_obj
        else:
            raise ValueError(
                'obj should have all keys ("type","path","params","value")'
            )

    def del_prod(self, obj, debug):
        if "type" in obj and "path" in obj:
            new_obj = copy.copy(obj)
            new_obj["path"] = self.get_path(obj.get("path"))
            new_obj["params"] = {}
            new_obj["file"] = None
            new_obj["value"] = None
            new_obj["status"] = t_delete
            try:
                if debug:
                    print(f'{new_obj["status"]} ==> {new_obj}')
                r = requests.post(f"{n_env.api}/{t_job}", json=new_obj)
                r.raise_for_status()
                res = r.json()
                if debug:
                    print(f'{res["status"]} ==> {res}')
            except requests.exceptions.ConnectionError as err:
                print(self.__error_busy, err)
                raise
            except requests.HTTPError as err:
                print(self.__error_reject, err)
                raise
            return new_obj
        else:
            raise ValueError('obj should have keys ("type","path")')



from naas.types import t_notebook, t_scheduler, t_error, t_output
from .env_var import n_env
from nbconvert import HTMLExporter
from sanic import response
import papermill as pm
import traceback
import mimetypes
import json
import time
import bs4
import csv
import os
import io
import datetime
import shutil

kern_manager = None
mime_html = "text/html"
mime_json = "application/json"
mime_jpeg = "image/jpeg"
mime_png = "image/png"
mime_svg = "image/svg+xml"
mime_list = [mime_html, mime_jpeg, mime_png, mime_svg]

try:
    from enterprise_gateway.services.kernels.remotemanager import RemoteKernelManager

    kern_manager = RemoteKernelManager
except ImportError:
    pass


class Notebooks:
    __logger = None
    __notif = None
    __html_exporter = None

    def __init__(self, logger, notif=None):
        self.__logger = logger
        self.__notif = notif
        self.__html_exporter = HTMLExporter()
        self.__html_exporter.template_name = "lab"

    def response(self, uid, filepath, res, duration, params):
        next_url = params.get("next_url", None)
        if next_url is not None and "https://" not in next_url:
            self.__logger.error(
                {
                    "id": uid,
                    "type": t_notebook,
                    "status": "next_url",
                    "filepath": filepath,
                    "url": next_url,
                    "error": "url not in right format",
                }
            )
        if next_url is not None and "https" in next_url:
            self.__logger.info(
                {
                    "id": uid,
                    "type": t_notebook,
                    "status": "next_url",
                    "filepath": filepath,
                    "url": next_url,
                }
            )
            return response.redirect(next_url)
        else:
            res_data = self.__get_res(res, filepath)
            if res_data and res_data.get("type"):
                file_name = os.path.basename(filepath)
                ext = mimetypes.guess_extension(res_data.get("type"), strict=True)
                inline = params.get("inline", False)
                headers = dict()
                new_file_name = f'{file_name.split(".")[0]}{ext}'
                if ext and not inline:
                    headers[
                        "Content-Disposition"
                    ] = f'attachment; filename="{new_file_name}"'

                async def streaming_fn(res):
                    # await res.write(str(res_data.get("data")))
                    await res.write(str(res_data.get("data")).encode("utf-8"))

                return response.stream(
                    streaming_fn,
                    headers=headers,
                    content_type=mimetypes.guess_type(new_file_name)[0],
                )
            else:
                return response.json({"id": uid, "status": "Done", "time": duration})

    def __convert_csv(self, data):
        soup = bs4.BeautifulSoup(data, features="html5lib")
        output = []
        for table_num, table in enumerate(soup.find_all("table")):
            csv_string = io.StringIO()
            csv_writer = csv.writer(csv_string, delimiter=";", quoting=csv.QUOTE_ALL)
            for tr in table.find_all("tr"):
                row = [
                    "".join(cell.stripped_strings) for cell in tr.find_all(["td", "th"])
                ]
                csv_writer.writerow(row)
            table_attrs = dict(num=table_num)
            output.append((csv_string.getvalue(), table_attrs))
        return output[0][0]

    def get_out_path(self, path):
        filename = os.path.basename(path)
        dirname = os.path.dirname(path)
        out_path = os.path.join(dirname, f"{t_output}_{filename}")
        return out_path

    def __nb_render(self, filepath):
        result_type = None
        result = None
        try:
            result_type = mime_html
            file_filepath_out = self.get_out_path(filepath)
            (result, ressources) = self.__html_exporter.from_filename(file_filepath_out)
        except FileNotFoundError:
            tb = traceback.format_exc()
            result_type = mime_json
            result = {
                "error": "output file not found",
                "trace": tb,
            }
        return result_type, result

    def __nb_file(self, meta, data):
        result_type = None
        result = None
        try:
            result_type = meta.get("naas_type")
            path = data.get(mime_json).get("path")
            with open(path, "r") as f:
                result = f.read()
                f.close()
        except FileNotFoundError:  # noqa: E722
            result_type = mime_json
            result = {"error": "file not found"}
        return result_type, result

    def __check_output(self, output, filepath):
        metadata = output.get("metadata", [])
        data = output.get("data", dict())
        meta_filtered = list(
            filter(lambda meta: metadata[meta].get("naas_api"), metadata)
        )
        for meta in meta_filtered:
            if (
                data.get("text/markdown")
                and metadata[meta].get("naas_type") == t_notebook
            ):
                return self.__nb_render(filepath)
            elif data.get(mime_json) and metadata[meta].get("naas_type"):
                return self.__nb_file(metadata[meta], data)
            elif data.get(mime_html) and metadata[meta].get("naas_type") == "csv":
                return "text/csv", self.__convert_csv(data.get(mime_html))
            elif data.get(mime_json):
                return mime_json, json.dumps(data.get(mime_json))
            else:
                result_type = next((i for i in mime_list if data.get(i)), None)
                return result_type, data.get(result_type)
        return None, None

    def __get_res(self, res, filepath):
        cells = res.get("cells")
        for cell in cells:
            outputs = cell.get("outputs", [])
            for output in outputs:
                (result_type, result) = self.__check_output(output, filepath)
                if result is not None and result_type is not None:
                    return {"type": result_type, "data": result}
        return None

    def __keep_out_history(self, file_filepath_out):
        if os.path.exists(file_filepath_out):
            try:
                out_finename = os.path.basename(file_filepath_out)
                out_dir = os.path.dirname(file_filepath_out)
                history_filename = f'{datetime.datetime.now().strftime("%Y%m%d%H%M%S%f")}_{out_finename}'
                history_path = os.path.join(out_dir, history_filename)
                shutil.copy(file_filepath_out, history_path)
            except:  # noqa: E722
                pass

    def __pm_exec(self, uid, file_dirpath, file_filepath, file_filepath_out, params):
        res = None
        if kern_manager:
            res = pm.execute_notebook(
                input_path=file_filepath,
                output_path=file_filepath_out,
                progress_bar=False,
                cwd=file_dirpath,
                parameters=params,
                kernel_manager_class=kern_manager,
            )
        else:
            res = pm.execute_notebook(
                input_path=file_filepath,
                output_path=file_filepath_out,
                progress_bar=False,
                cwd=file_dirpath,
                parameters=params,
            )
        if not res:
            res = {"uid": uid, "error": "Unknow error", "duration": 0}
        self.__keep_out_history(file_filepath_out)
        return res

    def __send_notification(self, uid, res, file_filepath, current_type, value, params):
        notif_down = params.get("notif_down", None)
        notif_up = params.get("notif_up", None)
        small_path = file_filepath.replace(n_env.server_root, "")
        if res.get("error"):
            if n_env.user is not None:
                self.__notif.send_status(
                    uid, "down", n_env.user, small_path, current_type, value
                )
            if notif_down and self.__notif:
                self.__notif.send_status(
                    uid, "down", notif_down, small_path, current_type, value
                )
        elif notif_up and current_type == t_scheduler and self.__notif:
            self.__notif.send_status(
                uid, "up", notif_down, small_path, current_type, value
            )
        elif notif_up and self.__notif:
            self.__notif.send_status(uid, "up", notif_up, small_path, current_type)

    def __get_output_path(self, file_filepath):
        file_dirpath = os.path.dirname(file_filepath)
        file_filename = os.path.basename(file_filepath)
        file_filepath_out = os.path.join(file_dirpath, f"out_{file_filename}")
        return file_filepath_out

    async def exec(self, uid, job):
        value = job.get("value", None)
        current_type = job.get("type", None)
        file_filepath = job.get("path")
        if not os.path.exists(file_filepath):
            err = "file not found"
            self.__logger.error(
                {
                    "id": uid,
                    "type": "filepath",
                    "status": t_error,
                    "filepath": file_filepath,
                    "error": err,
                }
            )
            return {"error": err, "duration": 0}
        file_dirpath = os.path.dirname(file_filepath)
        file_filepath_out = self.__get_output_path(file_filepath)
        params = job.get("params", dict())
        params["naas_uid"] = uid
        params["naas_runtime"] = datetime.datetime.now().strftime("%Y%m%d%H%M%S%f")
        params["naas_env"] = "PRODUCTION"
        start_time = time.time()
        res = None
        try:
            res = self.__pm_exec(
                uid, file_dirpath, file_filepath, file_filepath_out, params
            )
        except pm.PapermillExecutionError as err:
            tb = traceback.format_exc()
            res = {"error": err, "traceback": str(tb)}
            self.__logger.error(
                {
                    "id": uid,
                    "type": "PapermillExecutionError",
                    "status": t_error,
                    "filepath": file_filepath,
                    "output_filepath": file_filepath_out,
                    "error": str(err),
                }
            )
        except pm.PapermillException as err:
            tb = traceback.format_exc()
            res = {"error": err, "traceback": str(tb)}
            self.__logger.error(
                {
                    "id": uid,
                    "type": "Exception",
                    "status": t_error,
                    "filepath": file_filepath,
                    "output_filepath": file_filepath_out,
                    "error": err,
                    "traceback": str(tb),
                }
            )

        except RuntimeError as err:
            tb = traceback.format_exc()
            res = {"error": err, "traceback": str(tb)}
            self.__logger.error(
                {
                    "id": uid,
                    "type": "Exception",
                    "status": t_error,
                    "filepath": file_filepath,
                    "output_filepath": file_filepath_out,
                    "error": err,
                    "traceback": str(tb),
                }
            )
        except:  # noqa: E722
            tb = traceback.format_exc()
            res = {"error": "Unknow error", "traceback": str(tb)}
            self.__logger.error(
                {
                    "id": uid,
                    "type": "Exception",
                    "status": t_error,
                    "filepath": file_filepath,
                    "output_filepath": file_filepath_out,
                    "error": res.get("error"),
                    "traceback": str(tb),
                }
            )
        res["duration"] = time.time() - start_time
        try:
            self.__send_notification(
                uid, res, file_filepath_out, current_type, value, params
            )
        except ConnectionError as err:
            self.__logger.error(
                {
                    "id": uid,
                    "type": "Exception",
                    "status": t_error,
                    "filepath": file_filepath,
                    "output_filepath": file_filepath_out,
                    "error": "Cannot send notification",
                    "traceback": str(err),
                }
            )
        return res



from naas.types import t_main, t_notebook, t_scheduler, t_asset, t_job, t_secret
from sentry_sdk.integrations.sanic import SanicIntegration
from .controllers.scheduler import SchedulerController
from .controllers.assets import AssetsController
from .controllers.secret import SecretController
from .controllers.notebooks import NbController
from .controllers.jobs import JobsController
from .controllers.logs import LogsController
from sanic_openapi import swagger_blueprint
from .controllers.env import EnvController
from .notifications import Notifications
from .proxy import escape_kubernet
from .scheduler import Scheduler
from .notebooks import Notebooks
from .env_var import n_env
from .logger import Logger
from .jobs import Jobs
from .secret import Secret
from sanic import Sanic
import sentry_sdk
import asyncio
import getpass
import uuid
import os
import sys
import errno
import nest_asyncio

# TODO remove this fix when papermill support uvloop of Sanic support option to don't use uvloop
asyncio.set_event_loop_policy(None)
nest_asyncio.apply()

__version__ = "0.23.12"


class Runner:
    __naas_folder = ".naas"
    # Declare path variable
    __path_lib_files = os.path.dirname(os.path.abspath(__file__))
    __html_files = "html"
    __manager_index = "manager.html"
    __app = None
    __nb = None
    __notif = None
    __jobs = None
    __scheduler = None
    __logger = None

    def __init__(self):
        self.__path_naas_files = os.path.join(n_env.server_root, self.__naas_folder)
        self.__path_html_files = os.path.join(self.__path_lib_files, self.__html_files)
        self.__path_manager_index = os.path.join(
            self.__path_html_files, self.__manager_index
        )

    def __main(self, debug=True):
        self.init_app()
        uid = str(uuid.uuid4())
        self.__logger.info(
            {"id": uid, "type": t_main, "filepath": "runner", "status": "start API"}
        )
        self.__app.run(
            host="0.0.0.0", port=n_env.api_port, debug=debug, access_log=debug
        )

    async def initialize_before_start(self, app, loop):
        if self.__jobs is None:
            self.__jobs = Jobs(self.__logger)
            self.__secret = Secret(self.__logger)
            self.__nb = Notebooks(self.__logger, self.__notif)
            self.__app.add_route(
                NbController.as_view(self.__logger, self.__jobs, self.__nb),
                f"/{t_notebook}/<token>",
            )
            if n_env.scheduler:
                self.__scheduler = Scheduler(
                    self.__logger, self.__jobs, self.__nb, loop
                )
            self.__app.add_route(
                SchedulerController.as_view(self.__scheduler, self.__logger),
                f"/{t_scheduler}/<mode>",
            )
            self.__app.add_route(
                AssetsController.as_view(
                    self.__logger, self.__jobs, self.__path_lib_files
                ),
                f"/{t_asset}/<token>",
            )
            self.__app.add_route(
                JobsController.as_view(self.__logger, self.__jobs), f"/{t_job}"
            )
            self.__app.add_route(
                SecretController.as_view(self.__logger, self.__secret), f"/{t_secret}"
            )
            if n_env.scheduler:
                await self.__scheduler.start()

    async def initialize_before_stop(self, app, loop):
        if self.__nb is not None:
            self.__nb = None
        if n_env.scheduler and self.__scheduler is not None:
            await self.__scheduler.stop()
            self.__scheduler = None
        if self.__jobs is not None:
            self.__jobs = None

    def init_app(self):
        if not os.path.exists(self.__path_naas_files):
            try:
                print("Init Naas folder Jobs")
                os.makedirs(self.__path_naas_files)
            except OSError as exc:  # Guard against race condition
                print("__path_naas_files", self.__path_naas_files)
                if exc.errno != errno.EEXIST:
                    raise
            except Exception as e:
                print("Exception", e)
        self.__app = Sanic(__name__)
        self.__logger = Logger()
        self.__notif = Notifications(self.__logger)
        self.__app.register_listener(
            self.initialize_before_start, "before_server_start"
        )
        self.__app.register_listener(self.initialize_before_stop, "before_server_stop")
        self.__app.add_route(
            EnvController.as_view(
                self.__logger,
                n_env.user,
                n_env.hub_api,
                n_env.proxy_api,
                n_env.notif_api,
                n_env.tz,
                n_env.server_root,
            ),
            "/env",
        )
        self.__app.add_route(LogsController.as_view(self.__logger), "/log")
        self.__app.static("/", self.__path_manager_index, name="manager.html")
        self.__app.blueprint(swagger_blueprint)
        uid = str(uuid.uuid4())
        self.__logger.info(
            {"id": uid, "type": t_main, "filepath": "runner", "status": "init API"}
        )
        return self.__app

    def kill(self):
        self.__app.stop()

    def start(self, port=None, debug=False):
        user = getpass.getuser()
        if user != n_env.shell_user:
            raise ValueError(f"{user} not autorized, use {n_env.shell_user} instead")
        if port:
            n_env.api_port = port
        print("Start Runner", __version__)
        try:
            if n_env.sentry_dsn:
                sentry_sdk.init(
                    dsn=n_env.sentry_dsn,
                    traces_sample_rate=1.0,
                    environment=escape_kubernet(n_env.user),
                    integrations=[SanicIntegration()],
                )
                sentry_sdk.set_user({"email": n_env.user})
                with sentry_sdk.configure_scope() as scope:
                    scope.set_context("Naas", {"version": __version__})
            self.__main(debug)
        except KeyboardInterrupt:
            print("Shutdown server")
            sys.exit()



t_notebook = "notebook"
t_asset = "asset"
t_dependency = "dependency"
t_scheduler = "scheduler"
t_secret = "secret"

t_job = "job"
t_env = "env"
t_log = "log"

t_list_output = "list_output"
t_list_histo = "list_histo"
t_output = "output"
t_production = "prod"
t_histo = "history"
t_list = "list"
t_send = "send"
t_add = "installed"
t_delete = "delete"
t_update = "edited"
t_start = "started"
t_busy = "busy"
t_skip = "skiped"
t_error = "error"
t_health = "healthy"
t_main = "main"



from pathlib import Path
import os


class n_env:
    _api = None
    _version = None
    _remote_mode = False
    _api_port = None
    _notif_api = None
    _proxy_api = None
    _hub_api = None

    _server_root = None
    _shell_user = None

    _token = None
    _user = None
    _tz = None
    _sentry_dsn = None
    _scheduler = True
    _scheduler_interval = None
    _scheduler_job_max = None
    _scheduler_job_name = None
    _scheduler_timeout = None

    @property
    def api_port(self):
        return self._api_port or os.environ.get("NAAS_API_PORT", "5000")

    @api_port.setter
    def api_port(self, api_port: int):
        self._api_port = api_port

    @property
    def version(self):
        return self._version

    @version.setter
    def version(self, version: str):
        self._version = version

    @property
    def remote_mode(self):
        return self._remote_mode

    @remote_mode.setter
    def remote_mode(self, remote_mode):
        self._remote_mode = remote_mode

    @property
    def api(self):
        return self._api or os.environ.get(
            "NAAS_API",
            f"http://localhost:{self.api_port}"
            if not self.remote_mode
            else self.remote_api,
        )

    @api.setter
    def api(self, api):
        self._api = api

    @property
    def notif_api(self):
        return self._notif_api or os.environ.get(
            "NOTIFICATIONS_API", "https://notif.naas.ai"
        )

    @notif_api.setter
    def notif_api(self, notif_api):
        self._notif_api = notif_api

    @property
    def proxy_api(self):
        return self._proxy_api or os.environ.get("PROXY_API", "https://public.naas.ai")

    @proxy_api.setter
    def proxy_api(self, proxy_api):
        self._proxy_api = proxy_api

    @property
    def hub_api(self):
        return self._hub_api or os.environ.get("JUPYTERHUB_URL", "https://app.naas.ai")

    @hub_api.setter
    def hub_api(self, hub_api):
        self._hub_api = hub_api

    @property
    def server_root(self):
        return self._server_root or os.environ.get(
            "JUPYTER_SERVER_ROOT", str(Path.home())
        )

    @server_root.setter
    def server_root(self, server_root):
        self._server_root = server_root

    @property
    def shell_user(self):
        return self._shell_user or os.environ.get("NB_USER", "ftp")

    @shell_user.setter
    def shell_user(self, shell_user):
        self._shell_user = shell_user

    @property
    def remote_api(self):
        return f"{self.proxy_api}/runner"

    @property
    def token(self):
        return self._token or os.environ.get("JUPYTERHUB_API_TOKEN", "")

    @token.setter
    def token(self, token):
        self._token = token

    @property
    def user(self):
        return self._user or os.environ.get("JUPYTERHUB_USER", "")

    @user.setter
    def user(self, user):
        self._user = user

    @property
    def tz(self):
        return self._tz or os.environ.get("TZ", "Europe/Paris")

    @tz.setter
    def tz(self, tz):
        self._tz = tz

    @property
    def sentry_dsn(self):
        return self._sentry_dsn or os.environ.get("NAAS_SENTRY_DSN", None)

    @sentry_dsn.setter
    def sentry_dsn(self, sentry_dsn):
        self._sentry_dsn = sentry_dsn

    @property
    def scheduler(self):
        return bool(self._scheduler)

    @scheduler.setter
    def scheduler(self, scheduler):
        self._scheduler = bool(scheduler)

    @property
    def scheduler_interval(self):
        return int(
            self._scheduler_interval or os.environ.get("NAAS_SCHEDULER_INTERVAL", "60")
        )

    @scheduler_interval.setter
    def scheduler_interval(self, scheduler_interval: int):
        self._scheduler_interval = scheduler_interval

    @property
    def scheduler_job_max(self):
        return int(
            self._scheduler_job_max or os.environ.get("NAAS_SCHEDULER_JOB_MAX", "60")
        )

    @scheduler_job_max.setter
    def scheduler_job_max(self, scheduler_job_max: int):
        self._scheduler_job_max = scheduler_job_max

    @property
    def scheduler_job_name(self):
        return self._scheduler_job_name or os.environ.get(
            "NAAS_SCHEDULER_JOB_NAME", "_scheduler_job"
        )

    @scheduler_job_name.setter
    def scheduler_job_name(self, scheduler_job_name):
        self._scheduler_job_name = scheduler_job_name

    @property
    def scheduler_timeout(self):
        return int(
            self._scheduler_timeout or os.environ.get("NAAS_SCHEDULER_TIMEOUT", "3600")
        )

    @scheduler_timeout.setter
    def scheduler_timeout(self, scheduler_timeout: int):
        self._scheduler_timeout = scheduler_timeout


n_env = n_env()
